/*
  LIB16.FMT
  Archive support plugin for OMF-Library
  Second-level plugin for FAR Manager

  Valentin Skirdin
    vskirdin@mail.ru

  ver 1.0  15.07.1999

*/
/*
  Used document:
    "RELOCATABLE OBJECT MODULE FORMAT"
       http://www.neutralzone.org/home/faqsys/docs/ss0288_1.txt

    "the format of object files"
       http://www.neutralzone.org/home/faqsys/docs/object.fmt

  Documents also base in this archive
*/
#include <windows.h>
#include <string.h>
#include <dos.h>
#include "../fmt.hpp"
#include "plugin.hpp"


#define THEADR  0x80   // Translator Header Record
#define MODEND  0x8A   // Module End Record
#define MODEND2 0x8B   // Module End Record

#define RIDATA	0x74   // Relocatable Iterated Data Record

#define EXTDEF  0x8C   // External Names Definition Record
#define PUBDEF  0x90   // Public Names Definition Record
#define PUBDEF2 0x91   // Public Names Definition Record

/*
Record Format

All object records conform to the following format:

                          <------Record Length in Bytes----->
   1         2            <variable>    1
   Record    Record       Record        Checksum or 0
   Type      Length       Contents

*/
struct recHeader{
  BYTE  RecordType;   // identifies the type of object record
  WORD  RecordLength; // object record in bytes (excluding the bytes
                      // in the RecordType and RecordLength fields)
  // An entire record occupies 3 bytes plus the number of bytes in
  // the RecordLength field.
};

/*
Library Header Record (<n> bytes)

   1     2                    4           2            1       <n - 10>
   Type  Record Length        Dictionary  Dictionary   Flags   Padding
                              Offest      Size
   (F0H)  (Page Size Minus 3)             in Blocks
*/
struct LibHeader{
  struct recHeader F0;
  DWORD  DictOffset;
  WORD   DictSize;
  BYTE   flags;
};

/*
㫨  ⥪ ᥣ 稭  砫 ࠭.
 ࠭ ।  Record Length +3 (⠪
ࠧ   ᥣ  筮  ࠭).
 稭  ࠭筮 ࠧ   2  <n>,
 <n>  >= 4  <= 15.
*/
#define PAGE_SIZE	(libHeader.F0.RecordLength+3)


static char tmp[512];
static struct LibHeader libHeader;
static HANDLE ArcHandle;
static DWORD CurPos,NextPosition,FileSize;

        
BOOL WINAPI _export IsArchive(char */*Name*/,const unsigned char *Data,int /*DataSize*/)
{
  libHeader=*(struct LibHeader *)Data;
  // Library Header Record?
  if(libHeader.F0.RecordType == 0xF0)
    return(TRUE);
  return(FALSE);
}

BOOL WINAPI _export OpenArchive(char *Name,int *Type)
{
  ArcHandle=CreateFile(Name,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,
                       NULL,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,NULL);
  if (ArcHandle==INVALID_HANDLE_VALUE)
    return(FALSE);
  *Type=0;

  FileSize=GetFileSize(ArcHandle,NULL);

  // ।⢥   ᫥  ꥪ
  //     ⥪.
  //   - Microsoft OMF (稭  LHEADR  
  //   稢 MODEND ). 
  NextPosition=SetFilePointer(ArcHandle,libHeader.F0.RecordLength+
                              sizeof(recHeader),NULL,FILE_BEGIN);
  return(TRUE);
}

int WINAPI _export GetArcItem(struct PluginPanelItem *Item,struct ArcItemInfo *Info)
{
  DWORD ReadSize,pos,
        begin_pos=0,end_pos=0; //  ᫥ ࠧ OBJ-
  int name_len;
  recHeader rh;

  NextPosition=SetFilePointer(ArcHandle,NextPosition,NULL,FILE_BEGIN);
  if (NextPosition == 0xFFFFFFFF)
   return(GETARC_READERROR);
  if (NextPosition > FileSize)
   return(GETARC_UNEXPEOF);

  // ६  ᫥饣 䠩
  while(1)
  {
    if (!ReadFile(ArcHandle,&rh,sizeof(recHeader),&ReadSize,NULL))
      return(GETARC_READERROR);
    if (ReadSize < sizeof(recHeader))
      return(GETARC_EOF);

    // Library End Record?
    if(rh.RecordType == 0xF1)
      return(GETARC_EOF);

    //   ᫥ 樨 ᫥饣 䠩.
    // 筮 ᤥ ""... :-)  ⮤ 몠.
    if(rh.RecordType == RIDATA)
    {
      if((NextPosition%PAGE_SIZE))
	NextPosition=NextPosition+
                     PAGE_SIZE-
                     (NextPosition%PAGE_SIZE);
      else
	NextPosition+=PAGE_SIZE;
      SetFilePointer(ArcHandle,NextPosition,NULL,FILE_BEGIN);
      continue;
    }

    // Translator Header Record?
    if(rh.RecordType == THEADR)
    {
      /* The THEADR record contains the name of the object module.
         Record Format

         1    2           1            <-String Length->  1
         80   Record      String       Name String        Checksum
              Length      Length
      */
      if (!ReadFile(ArcHandle,tmp,rh.RecordLength,&ReadSize,NULL)
         || (WORD)ReadSize!=rh.RecordLength) //  check sum
        return(GETARC_READERROR);
      name_len=rh.RecordLength;

      begin_pos=NextPosition;
      NextPosition+=sizeof(recHeader)+rh.RecordLength; //??

      // Find Module End Record
      while(1)
      {
        if (!ReadFile(ArcHandle,&rh,sizeof(recHeader),&ReadSize,NULL))
          return(GETARC_READERROR);
        if (ReadSize < sizeof(recHeader))
          return(GETARC_EOF);

        // Library End Record?
	if(rh.RecordType == 0xF1)
	   return(GETARC_EOF);

        // Module End Record?
        if(rh.RecordType == MODEND || rh.RecordType == MODEND2)
        {
          /*
            MODEND record denotes the end of an object module
          */
          end_pos=(NextPosition+=sizeof(recHeader)+rh.RecordLength)-1; //??

	  if((end_pos%PAGE_SIZE))
	    NextPosition=end_pos+
                         PAGE_SIZE-
                         (end_pos%PAGE_SIZE);
          break;
        }
        else
          NextPosition=SetFilePointer(ArcHandle,NextPosition+
                                                sizeof(recHeader)+
                                                rh.RecordLength,
                                      NULL,FILE_BEGIN);
      }
      // ࠭  OBJ-...
      //   ᯮ짮:
      //    88H LIBMOD--LIBRARY MODULE NAME RECORD (COMMENT CLASS A3)
      tmp[name_len-1]=0;
      memcpy(Item->FindData.cFileName,&tmp[1],name_len-1);
      char ch;
      int i;
      for(i=0; (ch=Item->FindData.cFileName[i]) != 0; ++i)
        if(ch == '.')
        {
          Item->FindData.cFileName[i]=0;
          break;
        }

      // ⢥   ᦠ ...
      Item->PackSize=Item->FindData.nFileSizeLow=
                end_pos - begin_pos-
                6 - strlen(Item->FindData.cFileName); //?
      //        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      //  ⮬,   LIB-䠩  OBJ-
      //   88H (COMMENT CLASS A3)
      //  TLIB  祭 䠩 㤠  

      strcat(Item->FindData.cFileName,".obj");
      Item->FindData.dwFileAttributes=FILE_ATTRIBUTE_NORMAL;
      Item->PackSizeHigh=Item->FindData.nFileSizeHigh=0;
      Item->FindData.dwReserved0=Item->FindData.dwReserved1=0;
      Item->UserData=Item->NumberOfLinks=Item->Flags=0;
      Item->Description=NULL;
      Item->CustomColumnData=NULL;
      Item->CustomColumnNumber=0;
      Item->Owner=NULL;

      // ਭ   ६   "த⥫" :-)
      GetFileTime(ArcHandle,
          &Item->FindData.ftCreationTime,
          &Item->FindData.ftLastAccessTime,
          &Item->FindData.ftLastWriteTime);

      //    㬠  ਫ  ᮢ뢠...
      //  ୮   ... TLIB ⮣   :-)
      Item->FindData.cAlternateFileName[0]=0;

      //  直 砩...
      for(i=0; i < 3; ++i)
        Item->Reserved[i]=0;
 
      Info->DictSize=0;
      return(GETARC_SUCCESS);
    }
    // Library End Record?
//    else if(rh.RecordType == 0xF1)
//      return(GETARC_EOF);
    else
      NextPosition=SetFilePointer(ArcHandle,NextPosition+
                                            sizeof(recHeader)+
                                            rh.RecordLength,
                                  NULL,FILE_BEGIN);
    if(NextPosition == 0xFFFFFFFF)
      return(GETARC_EOF);
  }
  return(GETARC_EOF);
}

BOOL WINAPI _export CloseArchive(struct ArcInfo * /*Info*/ )
{
  return(CloseHandle(ArcHandle));
}

BOOL WINAPI _export GetFormatName(int Type,char *FormatName,char *DefaultExt)
{
  if (Type==0)
  {
    strcpy(FormatName,"Library");
    strcpy(DefaultExt,"LIB");
    return(TRUE);
  }
  return(FALSE);
}


BOOL WINAPI _export GetDefaultCommands(int Type,int Command,char *Dest)
{
  if (Type==0)
  {
    static char *Commands[]={
      "",                //Extract               
      "tlib %%a *%%fW",                   //Extract without paths 
      "",                   //Test                  
      "tlib %%a -%%fW",                     //Delete                
      "",                            //Comment archive       
      "",                     //Comment files         
      "",                                     //Convert to SFX        
      "",                                                //Lock archive          
      "",                                                //Protect archive       
      "",                                       //Recover archive       
      "tlib %%a -+%%fW",        //Add files             
      "tlib %%a -+%%fW",        //Move files            
      "",  //Add files and folders 
      "",  //Move files and folders
      "*.*"                                              //"All files" mask      
    };
    if (Command<sizeof(Commands)/sizeof(Commands[0]))
    {
      strcpy(Dest,Commands[Command]);
      return(TRUE);
    }
  }
  return(FALSE);
}
